#ifndef __Ik_H
#define __Ik_H
/*
 * Copyright (c) 1996-1999 Picture Elements, Inc.
 *    Stephen Williams <steve@picturel.com>
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 */
#ifndef WIN32
#ident "$Id: Ik.h,v 1.38 1999/03/13 17:39:11 steve Exp $"
#endif

# include  <tcl.h>

/*
 * the Ik_Image represents a single ik_image object. The image kit
 * passes a pointer to this opaque type around to the functions and
 * drivers that handle images for the ik_image.
 */

typedef struct Ik_Image Ik_Image_t;
struct Ik_ImageFile;

/*
 * This is the canonical unit of image data. Unless raw forms are
 * specifically requested (or presented) IKIT library routines will
 * put one intensity per object of this type. In any case, warps are
 * ALWAYS specified in units of this type.
 */
typedef unsigned short ik_idata_t;

typedef struct ik_palette ik_palette_t;

/*
 * Find an ik_image with the given name, and return the token needed
 * to manipulate it with the operations below.
 */
struct Ik_Image* Ik_FindImage(const char*name);

/*
 * The image file, and possibly other sources, pass image data to the
 * ik_image with this function. The width and the height are the
 * dimensions in pixels, and the warp is the size of a line in
 * shorts. The pixel has some number of planes, each plane represented
 * by a 16 bit sample. So, grey would have 1 plane and each grey pixel
 * one byte, RGB color 3planes to make 6bytes per pixel.
 *
 * Once data is passed to the ik_image, it assumes the responsibility
 * of freeing it when it is no longer needed. The image data must
 * therefore be in malloc()ed memory.
 *
 * Ik_SetImage_fun
 *   This is the most general form of image set. It takes the
 *   dimensions of the image, image data and pointers to functions
 *   that can access the lines/pixels of the image.
 *
 * Ik_SetImage_pal
 *   This method supports creating an image that is stored in a
 *   palette16 form. That is, each ik_idata_t is an index into the
 *   palette that is the ik_palette_t member. Each pixel takes on
 *   ik_idata_t no matter how many planes there are.
 *
 * Ik_SetImage sets a 16bpi image with all the dimensions
 * specified. The palette is left uncalculated, because it may be
 * huge.
 *
 * Ik_SetImagex is obsolete.
 *
 * The Ik_GetImage retrieves all that information about the image. The
 * Ik_GetImageInfo is the same, but doesn't get the image data. The
 * latter along with the transform functions below are the preferred
 * interface. 
 */
extern void Ik_SetImage(struct Ik_Image*ik,
			unsigned short*data, unsigned planes,
			unsigned width, unsigned height, unsigned warp);
extern void Ik_SetImagex(struct Ik_Image*ik,
			 unsigned short*data, unsigned planes,
			 unsigned width, unsigned height,
			 unsigned warp, unsigned bpi);


extern void Ik_SetImage_pal(Ik_Image_t*ik, ik_idata_t*dp, unsigned width,
			    unsigned height, unsigned warp,
			    unsigned bpi, ik_palette_t*);

typedef void (*get_line_fun)(ik_idata_t*cp,
			     const struct Ik_Image*ik,
			     unsigned y);
typedef unsigned (*get_pixel_fun)(const struct Ik_Image*ik,
				  unsigned x, unsigned y,
				  ik_idata_t*cp, unsigned ncp);
extern void Ik_SetImage_fun(struct Ik_Image*ik,
			    ik_idata_t*data, unsigned planes,
			    unsigned width, unsigned height, unsigned warp,
			    get_line_fun lf, get_pixel_fun pf);

extern void Ik_GetImageInfo(const struct Ik_Image*ik, unsigned*planes,
			    unsigned*width, unsigned*height);

/*
 * The Ik_SetImage_fun function takes one of these function sets as a
 * means of extracting pixels or lines out of the image data.
 *
 * ik_get_line_16bpi
 * ik_get_pixel_16bpi
 *   This is a format for accurate continuous tone images. Eash
 *   ik_idata_t stores a single 16bit intensity.
 *
 * ik_get_line_8bpi
 * ik_get_pixel_8bpi
 *   This is a format where continuous tone images are store as 8bpi
 *   values. An ik_idata_t* is cast to unsigned char*.
 *
 * ik_get_line_bin16
 * ik_get_pixel_bin16
 *   This is a format for storing binary images, where each ik_idata_t
 *   datum actually contains 16 pixels. The pixels are stored in the
 *   datum with the high order bit the left most in the image.
 */

extern void ik_get_line_16bpi(ik_idata_t*dp, const Ik_Image_t*ik,
			      unsigned y);
extern unsigned ik_get_pixel_16bpi(const Ik_Image_t*ik,
				   unsigned x, unsigned y,
				   ik_idata_t*cp, unsigned ncp);

extern void ik_get_line_8bpi(ik_idata_t*dp, const Ik_Image_t*ik,
			     unsigned y);
extern unsigned ik_get_pixel_8bpi(const Ik_Image_t*ik,
				  unsigned x, unsigned y,
				  ik_idata_t*cp, unsigned ncp);

extern void ik_get_line_bin16(ik_idata_t*dp, const struct Ik_Image*ik,
			      unsigned y);
extern unsigned ik_get_pixel_bin16(const struct Ik_Image*ik,
				   unsigned x, unsigned y,
				   ik_idata_t*cp, unsigned ncp);

/*
 * These are types of transforms that can be applied to images. In
 * every case, the source image is clipped by the destination image
 * which is placed over the source image by the off_x and off_y
 * parameters. If there are two or more source images, they must have
 * the same dimensions (they can have different depths).
 *
 * In all cases, if the dst image is 0, it iwll not be tracked and the
 * off_x and off_y parameters will be ignored. The scan dimensions
 * will come from the source image(s) in this case.
 *
 * The source images will always be presented as 16bpi, so that is the
 * only density the processing function may function. The destination
 * image line points to the beginning of the data and it is up to the
 * processing function to get this right. This is a very reasonable
 * requirement as the processing usually know a priori what the
 * density should be.
 *
 * The RAW functions will present the data in different formats. In
 * this cate, the bpi field will give the format used by the source.
 */

struct ik_context {
      unsigned s_width[2];  /* Width of the source image */
      unsigned s_planes[2]; /* planes in the source image */
};

typedef void (*ik_1op_f)(const ik_idata_t*s,
			 const struct ik_context*ctx, void*cd);

/* Iterate all over lines, one source, no surrounding context. */
extern void Ik_line_1op(struct Ik_Image*src,
			unsigned off_x, unsigned off_y,
			ik_1op_f fun, void*cd);

typedef void (*ik_2op_f)(const ik_idata_t*s1, const ik_idata_t*s2,
			 const struct ik_context*ctx, void*cd);

extern void Ik_line_2op(struct Ik_Image*src1, struct Ik_Image*src2,
			unsigned x, unsigned y, ik_2op_f fun, void*cd);


/*
 * These support operators that have some context from the source
 * images. The response to the callback is an array of line
 * pointers. The line being processed is at s[0]. Previous context is
 * at s[-1] etc, and later lines at s[1] etc. If the source image runs
 * out before or after, the s[-1] or s[1] pointer will as appropriate
 * be nil. It is up to the user to do the right thing.
 */

typedef void (*ik_n1op_f)(const ik_idata_t*const*s,
			  const struct ik_context*ctx, void*cd);

/* Iterate over lines, 1 source, up and down context. */
extern void Ik_nline_1op(struct Ik_Image*src1,
			 unsigned off_x, unsigned off_y,
			 unsigned con_up, unsigned con_do,
			 ik_n1op_f fun, void*cd);

/*
 * Every image has a PALETTE, which is an ordinal ordering of all the
 * colors in the image. The palette entry for a color is a single
 * integer that can be used to represent the color in the context of
 * the image.
 *
 * ik_pdatum_t
 *   This type in the integer representation of a color. This is used
 *   to represent a color relative to a palette.
 *
 * Ik_palette_size
 *   This returns the number of colors in the palette for the image.
 *
 * Ik_find_palette
 *   This function locates a color in the palette. The result is the
 *   palette entry for the color.
 *
 * Ik_palette_color
 *   This method returns the color information for the palette
 *   entry. The result is the color intensities.
 *
 * Ik_palette_count
 *   This returns the number of times the color appears in the
 *   image. The first call to this may be expensive, as it may need to
 *   generate a counts table.
 */

typedef unsigned ik_pdatum_t;

extern ik_palette_t* Ik_set_palette(unsigned size, unsigned planes,
				    ik_idata_t*data, unsigned*counts);

extern void Ik_attach_palette(Ik_Image_t*ik, ik_palette_t*pp);

extern unsigned Ik_palette_size(Ik_Image_t*);

extern ik_pdatum_t Ik_find_palette(Ik_Image_t*, const ik_idata_t*dp);

extern const ik_idata_t* Ik_palette_color(Ik_Image_t*, ik_pdatum_t);
extern unsigned Ik_palette_count(Ik_Image_t*, ik_pdatum_t);


/*
 * This function gets the value of a specific pixel in the image. The
 * caller must supply memory for the intensities (as an array of
 * shorts) and pass the size of that memory. The call will then return
 * the number of intensities actually included in the pixel.
 */
extern unsigned Ik_GetPixel(struct Ik_Image*ik, unsigned x, unsigned y,
			    unsigned short*cp, unsigned ncp);


/*
 * Use this method to add a new image file type to the ik_image. The
 * type pointer contains all the information necessary for the
 * ik_image to do what it needs to do to get or store the image.
 */
extern void Ik_CreateImageFile(struct Ik_ImageFile*type);
extern struct Ik_ImageFile*Ik_GetImageFile(const char*name);

/*
 * The read function for the ImageFile takes an ik pointer and a list
 * of arguments and reads the file. the arguments come from the image
 * create command line and are used to identify the file, and other
 * things.
 */
typedef int (*Ik_ImageFileOp)(Tcl_Interp*interp,
			      struct Ik_Image*ik,
			      int argc, char*argv[]);

/*
 * This structure represents a handle to an image file type. The
 * Ik_CreateImageFile above uses this structure to carry the methods
 * to the ikit core.
 *
 * the ImageFile interface is also used to create pseudo-sources, such
 * as different kinds of images, image processing operators,
 * etc. Every image type should support the read.
 *
 * Some image formats also support write, for example to write out a
 * TIFF file. In that case, the write method will be defined. Image
 * processing operators in general will NOT support write.
 */
struct Ik_ImageFile {
	/* Name of the file format */
      const char*type_name;

	/* Function to read the file image into memory. */
      Ik_ImageFileOp read;
      Ik_ImageFileOp write;

	/* Initialize the following to 0. ik_image uses them internally. */

      struct Ik_ImageFile*next;
};


/*
 * These are handy routines for reading matrices from Tcl array
 * variables. Every matrix has x and y dimensions, and the values are
 * stored in a single stretch of memory in raster order. If there is a
 * problem reading the matrix, the function returns 0 and an error
 * message is left in the interp.
 *
 * The array contains a specific set of elements to define the
 * matrix. If the array name is m, the elements are:
 *   m(x)   -- width of matrix
 *   m(y)   -- height of matrix
 *   m(a,b) -- value of matrix at coordinate a b, where a and b are
 *             unsigned integers where a<x and b<y.
 *
 * The matrix value returned is allocated by malloc() and must be
 * released by free() when no longer needed.
 */
extern long* matrix_l(Tcl_Interp*interp, const char*name,
		      unsigned*m_wid, unsigned*m_hei);


/*
 * $Log: Ik.h,v $
 * Revision 1.38  1999/03/13 17:39:11  steve
 *  Obsolete Ik_SetImagex
 *
 * Revision 1.37  1999/02/19 21:46:54  steve
 *  Prepare for 2.0 release.
 *
 * Revision 1.36  1999/02/15 00:39:37  steve
 *  Get rid of the Ik_GetImageBPI and any
 *  bpi members, as well as Ik_SetImageBinary.
 *
 * Revision 1.35  1999/02/14 22:43:27  steve
 *  Get ginary color sense right on display.
 *
 * Revision 1.34  1999/02/14 00:52:55  steve
 *  Make TIFF read binary images into a bin16 image.
 *
 * Revision 1.33  1999/02/03 16:46:57  steve
 *  Add the ndiffuse operation for collor
 *  diffusion, Rewrite palette handling,
 *  fix several file types to better use
 *  the palette information, rewrite the
 *  rendering fucntions to use palette
 *  information and white noise error diffusion,
 *  or otherwise be cleaner and faster,
 *  Add "visual" support in the ikit script.
 *
 * Revision 1.32  1998/05/25 23:17:03  steve
 *  Remove the dst paramter from Ik_line_2op.
 *
 * Revision 1.31  1998/05/25 19:08:45  steve
 *  Change the nline op prototype.
 *
 * Revision 1.30  1998/05/25 18:51:03  steve
 *  Ik_nline_op paramters changed.
 *
 * Revision 1.29  1998/05/23 05:27:11  steve
 *  Add bilinear upscaling.
 *
 * Revision 1.28  1998/05/19 04:31:27  steve
 *  Add generic image extraction behavior.
 *
 * Revision 1.27  1998/05/18 22:31:15  steve
 *  Get rid of Ik_line_1raw.
 *
 * Revision 1.26  1998/04/21 19:53:13  steve
 *  Add te plane laminate function.
 *
 * Revision 1.25  1998/03/19 02:54:15  steve
 *  Get rid of unused cruft.
 *
 * Revision 1.24  1998/03/12 01:06:57  steve
 *  Full-featured replication dimensions
 *
 * Revision 1.23  1998/01/12 20:46:29  steve
 *  Port to Windows NT, cleanup warnings.
 *
 * Revision 1.22  1997/10/11 01:37:11  steve
 *  Add PNG palette support.
 *
 * Revision 1.21  1997/09/19 18:06:45  steve
 *  Massive boost to colormap performance.
 *
 * Revision 1.20  1997/09/19 17:04:23  steve
 *  Add color quantization function.
 */
#endif
